"""
applications: concrete lightweight graph data structures
Corey Rayburn Yung <coreyrayburnyung@gmail.com>
Copyright 2020-2022, Corey Rayburn Yung
License: Apache-2.0

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

Contents:
     System
        
To Do:
    Complete Network which will use an adjacency matrix for internal storage.
    
"""
from __future__ import annotations
import collections
from collections.abc import (
    Collection, Hashable, MutableMapping, MutableSequence, Sequence, Set)
import dataclasses
from typing import Any, Optional, Type, TYPE_CHECKING, Union

import amos

from . import base
from . import check
from . import composites
from . import graphs
from . import report
from . import traits
from . import traverse
from . import workshop
    

@dataclasses.dataclass
class System(graphs.Adjacency, traits.Directed, traits.Fungible):
    """Directed graph with unweighted edges stored as an adjacency list.
    
    Args:
        contents (MutableMapping[Hashable, Set[Hashable]]): keys are nodes and 
            values are sets of nodes (or hashable representations of nodes). 
            Defaults to a defaultdict that has a set for its value format.
                  
    """  
    contents: MutableMapping[Hashable, Set[Hashable]] = (
        dataclasses.field(
            default_factory = lambda: collections.defaultdict(set)))
                                
    """ Properties """
    
    @property
    def endpoint(self) -> MutableSequence[Hashable]:
        """Returns the endpoints of the stored composite."""
        return report.get_endpoints_adjacency(item = self.contents)
                    
    @property
    def root(self) -> MutableSequence[Hashable]:
        """Returns the roots of the stored composite."""
        return report.get_roots_adjacency(item = self.contents)
                 
    """ Public Methods """

    def append(self, item: base.Graph) -> None:
        """Appends 'item' to the endpoints of the stored graph.

        Appending creates an edge between every endpoint of this instance's
        stored graph and the every root of 'item'.

        Args:
            item (base.Graph): another Graph, 
                an adjacency list, an edge list, an adjacency matrix, or one or
                more nodes.
            
        Raises:
            TypeError: if 'item' is neither a Graph, Adjacency, Edges, Matrix,
                or Collection[Hashable] type.
                
        """
        current_endpoints = self.endpoint
        if check.is_graph(item = item):
            self.merge(item = item)
            for endpoint in current_endpoints:
                for root in workshop.get_roots_adjacency(item = item):
                    self.connect((endpoint, root))
        elif check.is_node(item = item):
            self.add(item = item)
            for endpoint in current_endpoints:
                self.connect((endpoint, item))
        else:
            raise TypeError('item is not a recognized graph or node type')
        return                    

    def prepend(self, item: base.Graph) -> None:
        """Prepends 'item' to the roots of the stored graph.

        Prepending creates an edge between every endpoint of 'item' and every
        root of this instance;s stored graph.

        Args:
            item (base.Graph): another Graph, an adjacency list, an 
                edge list, an adjacency matrix, or one or more nodes.
            
        Raises:
            TypeError: if 'item' is neither a System, Adjacency, Edges, Matrix, 
                or Collection[Hashable] type.
                
        """
        current_roots = self.root
        if check.is_graph(item = item):
            self.merge(item = item)
            for root in current_roots:
                for endpoint in item.endpoint:
                    self.connect((endpoint, root))
        elif check.is_node(item = item):
            self.add(item = item)
            for root in current_roots:
                self.connect((item, root))
        else:
            raise TypeError('item is not a recognized graph or node type')
        return
    
    def walk(
        self, 
        start: Optional[Hashable] = None, 
        stop: Optional[Hashable] = None) -> Paths:
        """Returns all paths in graph from 'start' to 'stop'.
        
        Args:
            start (Hashable): node to start paths from.
            stop (Hashable): node to stop paths.
            
        Returns:
            Path: a list of possible paths (each path is a list nodes) from 
                'start' to 'stop'.
            
        """
        if start is None:
            roots = self.root
        else:
            roots = amos.listify(item = start)
        if stop is None:
            endpoints = self.endpoint
        else:
            endpoints = self.amos.listify(item = stop)
        all_paths = []
        for root in roots:
            for end in endpoints:
                paths = traverse.walk_adjacency(
                    item = self.contents, 
                    start = root, 
                    stop = end)
                if paths:
                    if all(isinstance(p, Hashable) for p in paths):
                        all_paths.append(paths)
                    else:
                        all_paths.extend(paths)
        return all_paths

 
# @dataclasses.dataclass
# class Network(Graph):
#     """composites class for undirected graphs with unweighted edges.
    
#     Graph stores a directed acyclic graph (DAG) as an adjacency list. Despite 
#     being called an adjacency "list," the typical and most efficient way to 
#     store one is using a python dict. a piles Graph inherits from a Dictionary 
#     in order to allow use of its extra functionality over a plain dict.
    
#     Graph supports '+' and '+=' to be used to join two piles Graph instances. A
#     properly formatted adjacency list could also be the added object.
    
#     Graph internally supports autovivification where a list is created as a 
#     value for a missing key. This means that a Graph need not inherit from 
#     defaultdict.
    
#     Args:
#         contents (Adjacency): an adjacency list where the keys are nodes and the 
#             values are nodes which the key is connected to. Defaults to an empty 
#             dict.
                  
#     """  
#     contents: Matrix = dataclasses.field(default_factory = dict)
    
#     """ Properties """

#     @property
#     def adjacency(self) -> Adjacency:
#         """Returns the stored graph as an adjacency list."""
#         return matrix_to_adjacency(item = self.contents)

#     @property
#     def breadths(self) -> Path:
#         """Returns all paths through the Graph using breadth-first search.
        
#         Returns:
#             Path: returns all paths from 'roots' to 'endpoints' in a list 
#                 of lists of nodes.
                
#         """
#         return self._find_all_paths(
#             starts = self.root, 
#             ends = self.endpoint,
#             depth_first = False)

#     @property
#     def depths(self) -> Path:
#         """Returns all paths through the Graph using depth-first search.
        
#         Returns:
#             Path: returns all paths from 'roots' to 'endpoints' in a list 
#                 of lists of nodes.
                
#         """
#         return self._find_all_paths(starts = self.root, 
#                                     ends = self.endpoint,
#                                     depth_first = True)
     
#     @property
#     def edges(self) -> Edges:
#         """Returns the stored graph as an edge list."""
#         return adjacency_to_edges(item = self.contents)

#     @property
#     def endpoints(self) -> list[Hashable]:
#         """Returns a list of endpoint nodes in the stored graph.."""
#         return [k for k in self.contents.keys() if not self.contents[k]]

#     @property
#     def matrix(self) -> Matrix:
#         """Returns the stored graph as an adjacency matrix."""
#         return adjacency_to_matrix(item = self.contents)
                      
#     @property
#     def nodes(self) -> dict[str, Hashable]:
#         """Returns a dict of node names as keys and nodes as values.
        
#         Because Graph allows various Hashable objects to be used as keys,
#         including the Collection[Hashable] class, there isn't an obvious way to access already
#         stored nodes. This property creates a new dict with str keys derived
#         from the nodes (looking first for a 'name' attribute) so that a user
#         can access a node. 
        
#         This property is not needed if the stored nodes are all strings.
        
#         Returns:
#             Dict[str, Hashable]: keys are the name or has of nodes and the 
#                 values are the nodes themselves.
            
#         """
#         return {self.trait.namify(item = n): n for n in self.contents.keys()}
  
#     @property
#     def roots(self) -> list[Hashable]:
#         """Returns root nodes in the stored graph..

#         Returns:
#             list[Hashable]: root nodes.
            
#         """
#         stops = list(itertools.chain(self.contents.values()))
#         return [k for k in self.contents.keys() if k not in stops]
    
#     """ Class Methods """
    
#     @classmethod
#     def create(cls, item: Union[Adjacency, Edges, Matrix]) -> Graph:
#         """Creates an instance of a Graph from 'item'.
        
#         Args:
#             item (Union[Adjacency, Edges, Matrix]): an adjacency list, 
#                 adjacency matrix, or edge list which can used to create the
#                 stored graph.
                
#         Returns:
#             Graph: a Graph instance created compositesd on 'item'.
                
#         """
#         if is_adjacency_list(item = item):
#             return cls.from_adjacency(adjacency = item)
#         elif is_adjacency_matrix(item = item):
#             return cls.from_matrix(matrix = item)
#         elif is_edge_list(item = item):
#             return cls.from_adjacency(edges = item)
#         else:
#             raise TypeError(
#                 f'create requires item to be an adjacency list, adjacency '
#                 f'matrix, or edge list')
           
#     @classmethod
#     def from_adjacency(cls, adjacency: Adjacency) -> Graph:
#         """Creates a Graph instance from an adjacency list.
        
#         'adjacency' should be formatted with nodes as keys and values as lists
#         of names of nodes to which the node in the key is connected.

#         Args:
#             adjacency (Adjacency): adjacency list used to 
#                 create a Graph instance.

#         Returns:
#             Graph: a Graph instance created compositesd on 'adjacent'.
              
#         """
#         return cls(contents = adjacency)
    
#     @classmethod
#     def from_edges(cls, edges: Edges) -> Graph:
#         """Creates a Graph instance from an edge list.

#         'edges' should be a list of tuples, where the first item in the tuple
#         is the node and the second item is the node (or name of node) to which
#         the first item is connected.
        
#         Args:
#             edges (Edges): Edge list used to create a Graph 
#                 instance.
                
#         Returns:
#             Graph: a Graph instance created compositesd on 'edges'.

#         """
#         return cls(contents = edges_to_adjacency(item = edges))
    
#     @classmethod
#     def from_matrix(cls, matrix: Matrix) -> Graph:
#         """Creates a Graph instance from an adjacency matrix.

#         Args:
#             matrix (Matrix): adjacency matrix used to create a Graph instance. 
#                 The values in the matrix should be 1 (indicating an edge) and 0 
#                 (indicating no edge).
 
#         Returns:
#             Graph: a Graph instance created compositesd on 'matrix'.
                        
#         """
#         return cls(contents = matrix_to_adjacency(item = matrix))
    
#     @classmethod
#     def from_path(cls, path: Path) -> Graph:
#         """Creates a Graph instance from a Path.

#         Args:
#             path (Path): serial path used to create a Graph
#                 instance.
 
#         Returns:
#             Graph: a Graph instance created compositesd on 'path'.
                        
#         """
#         return cls(contents = path_to_adjacency(item = path))
       
#     """ Public Methods """
    
#     def add(self, 
#             node: Hashable,
#             ancestors: Collection[Hashable] = None,
#             descendants: Collection[Hashable] = None) -> None:
#         """Adds 'node' to 'contents' with no corresponding edges.
        
#         Args:
#             node (Hashable): a node to add to the stored graph.
#             ancestors (Collection[Hashable]): node(s) from which node should be connected.
#             descendants (Collection[Hashable]): node(s) to which node should be connected.

#         """
#         if descendants is None:
#             self.contents[node] = []
#         elif descendants in self:
#             self.contents[node] = amos.iterify(item = descendants)
#         else:
#             missing = [n for n in descendants if n not in self.contents]
#             raise KeyError(f'descendants {missing} are not in the stored graph.')
#         if ancestors is not None:  
#             if (isinstance(ancestors, Hashable) and ancestors in self
#                     or (isinstance(ancestors, (list, tuple, set)) 
#                         and all(isinstance(n, Hashable) for n in ancestors)
#                         and all(n in self.contents for n in ancestors))):
#                 start = ancestors
#             elif (hasattr(self.__class__, ancestors) 
#                     and isinstance(getattr(type(self), ancestors), property)):
#                 start = getattr(self, ancestors)
#             else:
#                 missing = [n for n in ancestors if n not in self.contents]
#                 raise KeyError(f'ancestors {missing} are not in the stored graph.')
#             for starting in amos.iterify(item = start):
#                 if node not in [starting]:
#                     self.connect(start = starting, stop = node)                 
#         return 

#     def append(self, 
#                item: Union[Graph, Adjacency, Edges, Matrix, Collection[Hashable]]) -> None:
#         """Adds 'item' to this Graph.

#         Combining creates an edge between every endpoint of this instance's
#         Graph and the every root of 'item'.

#         Args:
#             item (Union[Graph, Adjacency, Edges, Matrix, Collection[Hashable]]): another 
#                 Graph to join with this one, an adjacency list, an edge list, an
#                 adjacency matrix, or Collection[Hashable].
            
#         Raises:
#             TypeError: if 'item' is neither a Graph, Adjacency, Edges, Matrix,
#                 or Collection[Hashable] type.
            
#         """
#         if isinstance(item, Graph):
#             if self.contents:
#                 current_endpoints = self.endpoint
#                 self.contents.update(item.contents)
#                 for endpoint in current_endpoints:
#                     for root in item.root:
#                         self.connect(start = endpoint, stop = root)
#             else:
#                 self.contents = item.contents
#         elif isinstance(item, Adjacency):
#             self.append(item = self.from_adjacency(adjacecny = item))
#         elif isinstance(item, Edges):
#             self.append(item = self.from_edges(edges = item))
#         elif isinstance(item, Matrix):
#             self.append(item = self.from_matrix(matrix = item))
#         elif isinstance(item, Collection[Hashable]):
#             if isinstance(item, (list, tuple, set)):
#                 new_graph = Graph()
#                 edges = more_itertools.windowed(item, 2)
#                 for edge_pair in edges:
#                     new_graph.add(node = edge_pair[0], descendants = edge_pair[1])
#                 self.append(item = new_graph)
#             else:
#                 self.add(node = item)
#         else:
#             raise TypeError(
#                 'item must be a Graph, Adjacency, Edges, Matrix, or Collection[Hashable] '
#                 'type')
#         return
  
#     def connect(self, start: Hashable, stop: Hashable) -> None:
#         """Adds an edge from 'start' to 'stop'.

#         Args:
#             start (Hashable): name of node for edge to start.
#             stop (Hashable): name of node for edge to stop.
            
#         Raises:
#             ValueError: if 'start' is the same as 'stop'.
            
#         """
#         if start == stop:
#             raise ValueError(
#                 'The start of an edge cannot be the same as the stop')
#         else:
#             if stop not in self.contents:
#                 self.add(node = stop)
#             if start not in self.contents:
#                 self.add(node = start)
#             if stop not in self.contents[start]:
#                 self.contents[start].append(self.trait.namify(item = stop))
#         return

#     def delete(self, node: Hashable) -> None:
#         """Deletes node from graph.
        
#         Args:
#             node (Hashable): node to delete from 'contents'.
        
#         Raises:
#             KeyError: if 'node' is not in 'contents'.
            
#         """
#         try:
#             del self.contents[node]
#         except KeyError:
#             raise KeyError(f'{node} does not exist in the graph')
#         self.contents = {
#             k: v.remove(node) for k, v in self.contents.items() if node in v}
#         return

#     def disconnect(self, start: Hashable, stop: Hashable) -> None:
#         """Deletes edge from graph.

#         Args:
#             start (Hashable): starting node for the edge to delete.
#             stop (Hashable): ending node for the edge to delete.
        
#         Raises:
#             KeyError: if 'start' is not a node in the stored graph..
#             ValueError: if 'stop' does not have an edge with 'start'.

#         """
#         try:
#             self.contents[start].remove(stop)
#         except KeyError:
#             raise KeyError(f'{start} does not exist in the graph')
#         except ValueError:
#             raise ValueError(f'{stop} is not connected to {start}')
#         return

#     def merge(self, item: Union[Graph, Adjacency, Edges, Matrix]) -> None:
#         """Adds 'item' to this Graph.

#         This method is roughly equivalent to a dict.update, just adding the
#         new keys and values to the existing graph. It converts the supported
#         formats to an adjacency list that is then added to the existing 
#         'contents'.
        
#         Args:
#             item (Union[Graph, Adjacency, Edges, Matrix]): another Graph to 
#                 add to this one, an adjacency list, an edge list, or an
#                 adjacency matrix.
            
#         Raises:
#             TypeError: if 'item' is neither a Graph, Adjacency, Edges, or 
#                 Matrix type.
            
#         """
#         if isinstance(item, Graph):
#             item = item.contents
#         elif isinstance(item, Adjacency):
#             pass
#         elif isinstance(item, Edges):
#             item = self.from_edges(edges = item).contents
#         elif isinstance(item, Matrix):
#             item = self.from_matrix(matrix = item).contents
#         else:
#             raise TypeError(
#                 'item must be a Graph, Adjacency, Edges, or Matrix type to '
#                 'update')
#         self.contents.update(item)
#         return
  
#     def subgraph(self, 
#                  include: Union[Any, Sequence[Any]] = None,
#                  exclude: Union[Any, Sequence[Any]] = None) -> Graph:
#         """Returns a new Graph without a subset of 'contents'.
        
#         All edges will be removed that include any nodes that are not part of
#         the new subgraph.
        
#         Any extra attributes that are part of a Graph (or a subclass) will be
#         maintained in the returned subgraph.

#         Args:
#             include (Union[Any, Sequence[Any]]): nodes which should be included
#                 with any applicable edges in the new subgraph.
#             exclude (Union[Any, Sequence[Any]]): nodes which should not be 
#                 included with any applicable edges in the new subgraph.

#         Returns:
#             Graph: with only key/value pairs with keys not in 'subset'.

#         """
#         if include is None and exclude is None:
#             raise ValueError('Either include or exclude must not be None')
#         else:
#             if include:
#                 excludables = [k for k in self.contents if k not in include]
#             else:
#                 excludables = []
#             excludables.extend([i for i in self.contents if i not in exclude])
#             new_graph = copy.deepcopy(self)
#             for node in amos.iterify(item = excludables):
#                 new_graph.delete(node = node)
#         return new_graph

#     def walk(self, 
#              start: Hashable, 
#              stop: Hashable, 
#              path: Path = None,
#              depth_first: bool = True) -> Path:
#         """Returns all paths in graph from 'start' to 'stop'.

#         The code here is adapted from: https://www.python.org/doc/essays/graphs/
        
#         Args:
#             start (Hashable): node to start paths from.
#             stop (Hashable): node to stop paths.
#             path (Path): a path from 'start' to 'stop'. Defaults to an 
#                 empty list. 

#         Returns:
#             Path: a list of possible paths (each path is a list 
#                 nodes) from 'start' to 'stop'.
            
#         """
#         if path is None:
#             path = []
#         path = path + [start]
#         if start == stop:
#             return [path]
#         if start not in self.contents:
#             return []
#         if depth_first:
#             method = self._depth_first_search
#         else:
#             method = self._breadth_first_search
#         paths = []
#         for node in self.contents[start]:
#             if node not in path:
#                 new_paths = self.walk(
#                     start = node, 
#                     stop = stop, 
#                     path = path,
#                     depth_first = depth_first)
#                 for new_path in new_paths:
#                     paths.append(new_path)
#         return paths

#     def _all_paths_bfs(self, start, stop):
#         """

#         """
#         if start == stop:
#             return [start]
#         visited = {start}
#         queue = collections.deque([(start, [])])
#         while queue:
#             current, path = queue.popleft()
#             visited.add(current)
#             for connected in self[current]:
#                 if connected == stop:
#                     return path + [current, connected]
#                 if connected in visited:
#                     continue
#                 queue.append((connected, path + [current]))
#                 visited.add(connected)   
#         return []

#     def _breadth_first_search(self, node: Hashable) -> Path:
#         """Returns a breadth first search path through the Graph.

#         Args:
#             node (Hashable): node to start the search from.

#         Returns:
#             Path: nodes in a path through the Graph.
            
#         """        
#         visited = set()
#         queue = [node]
#         while queue:
#             vertex = queue.pop(0)
#             if base. not in visited:
#                 visited.add(vertex)
#                 queue.extend(set(self[vertex]) - visited)
#         return list(visited)
       
#     def _depth_first_search(self, 
#         node: Hashable, 
#         visited: list[Hashable]) -> Path:
#         """Returns a depth first search path through the Graph.

#         Args:
#             node (Hashable): node to start the search from.
#             visited (list[Hashable]): list of visited nodes.

#         Returns:
#             Path: nodes in a path through the Graph.
            
#         """  
#         if node not in visited:
#             visited.append(node)
#             for edge in self[node]:
#                 self._depth_first_search(node = edge, visited = visited)
#         return visited
  
#     def _find_all_paths(self, 
#         starts: Union[Hashable, Sequence[Hashable]],
#         stops: Union[Hashable, Sequence[Hashable]],
#         depth_first: bool = True) -> Path:
#         """[summary]

#         Args:
#             start (Union[Hashable, Sequence[Hashable]]): starting points for 
#                 paths through the Graph.
#             ends (Union[Hashable, Sequence[Hashable]]): endpoints for paths 
#                 through the Graph.

#         Returns:
#             Path: list of all paths through the Graph from all
#                 'starts' to all 'ends'.
            
#         """
#         all_paths = []
#         for start in amos.iterify(item = starts):
#             for end in amos.iterify(item = stops):
#                 paths = self.walk(
#                     start = start, 
#                     stop = end,
#                     depth_first = depth_first)
#                 if paths:
#                     if all(isinstance(path, Hashable) for path in paths):
#                         all_paths.append(paths)
#                     else:
#                         all_paths.extend(paths)
#         return all_paths
            
#     """ Dunder Methods """

#     def __add__(self, other: Graph) -> None:
#         """Adds 'other' Graph to this Graph.

#         Adding another graph uses the 'merge' method. Read that method's 
#         docstring for further details about how the graphs are added 
#         together.
        
#         Args:
#             other (Graph): a second Graph to join with this one.
            
#         """
#         self.merge(graph = other)        
#         return

#     def __iadd__(self, other: Graph) -> None:
#         """Adds 'other' Graph to this Graph.

#         Adding another graph uses the 'merge' method. Read that method's 
#         docstring for further details about how the graphs are added 
#         together.
        
#         Args:
#             other (Graph): a second Graph to join with this one.
            
#         """
#         self.merge(graph = other)        
#         return

#     def __contains__(self, nodes: Collection[Hashable]) -> bool:
#         """[summary]

#         Args:
#             nodes (Collection[Hashable]): [description]

#         Returns:
#             bool: [description]
            
#         """
#         if isinstance(nodes, (list, tuple, set)):
#             return all(n in self.contents for n in nodes)
#         elif isinstance(nodes, Hashable):
#             return nodes in self.contents
#         else:
#             return False   
        
#     def __getitem__(self, key: Hashable) -> Any:
#         """Returns value for 'key' in 'contents'.

#         Args:
#             key (Hashable): key in 'contents' for which a value is sought.

#         Returns:
#             Any: value stored in 'contents'.

#         """
#         return self.contents[key]

#     def __setitem__(self, key: Hashable, value: Any) -> None:
#         """sets 'key' in 'contents' to 'value'.

#         Args:
#             key (Hashable): key to set in 'contents'.
#             value (Any): value to be paired with 'key' in 'contents'.

#         """
#         self.contents[key] = value
#         return

#     def __delitem__(self, key: Hashable) -> None:
#         """Deletes 'key' in 'contents'.

#         Args:
#             key (Hashable): key in 'contents' to delete the key/value pair.

#         """
#         del self.contents[key]
#         return

#     def __missing__(self) -> list:
#         """Returns an empty list when a key doesn't exist.

#         Returns:
#             list: an empty list.

#         """
#         return []
    
#     def __str__(self) -> str:
#         """Returns prettier summary of the Graph.

#         Returns:
#             str: a formatted str of class information and the contained 
#                 adjacency list.
            
#         """
#         new_line = '\n'
#         tab = '    '
#         summary = [f'{new_line}piles {self.__class__.__name__}']
#         summary.append('adjacency list:')
#         for node, edges in self.contents.items():
#             summary.append(f'{tab}{node}: {str(edges)}')
#         return new_line.join(summary) 

# Changer: Type[Any] = Callable[[Hashable], None]
# Finder: Type[Any] = Callable[[Hashable], Optional[Hashable]]




# @dataclasses.dataclass # type: ignore
# class Categorizer(Tree):
#     """composites class for an tree data structures.
        
#     Args:
#         contents (MutableSequence[Hashable]): list of stored Node 
#             instances (including other Trees). Defaults to an empty list.
#         name (Optional[str]): name of Tree node which should match a parent 
#             tree's key name corresponding to this Tree node. All nodes in a Tree
#             must have unique names. The name is used to make all Tree nodes 
#             hashable and capable of quick comparison. Defaults to None, but it
#             should not be left as None when added to a Tree.
#         parent (Optional[Tree]): parent Tree, if any. Defaults to None.
        
#     """
#     contents: MutableSequence[Hashable] = dataclasses.field(
#         default_factory = list)
#     name: Optional[str] = None
#     parent: Optional[Tree] = None 
    
#     """ Properties """
        
#     @property
#     def branches(self) -> list[Tree]:
#         """Returns all stored Tree nodes in a list."""
#         return self.nodes - self.leaves
    
#     @property
#     def children(self) -> dict[str, Hashable]:
#         """[summary]

#         Returns:
#             dict[str, Hashable]: [description]
#         """
#         return self.contents
    
#     @property
#     def is_leaf(self) -> bool:
#         """[summary]

#         Returns:
#             bool: [description]
#         """
#         return not self.children
    
#     @property
#     def is_root(self) -> bool:
#         """[summary]

#         Returns:
#             bool: [description]
#         """
#         return self.parent is None
    
#     @property
#     def leaves(self) -> list[Hashable]:
#         """Returns all stored leaf nodes in a list."""
#         matches = []
#         for node in self.nodes:
#             if not hasattr(node, 'is_leaf') or node.is_leaf:
#                 matches.append(node)
#         return matches
     
#     @property
#     def nodes(self) -> list[Hashable]:
#         """Returns all stored nodes in a list."""
#         return depth_first_search(tree = self.contents)

#     @property
#     def root(self) -> Tree:
#         """
#         """
#         composites = [n.is_root for n in self.nodes]
#         if len(composites) > 1:
#             raise ValueError('The tree is broken - it has more than 1 root')
#         elif len(composites) == 0:
#             raise ValueError('The tree is broken - it has no root')
#         else:
#             return composites[0]
    
#     """ Public Methods """
    
#     def add(
#         self, 
#         item: Union[Hashable, Sequence[Hashable]],
#         parent: Optional[str] = None) -> None:
#         """Adds node(s) in item to 'contents'.
        
#         In adding the node(s) to the stored tree, the 'parent' attribute for the
#         node(s) is set to this Tree instance.

#         Args:
#             item (Union[Hashable, Sequence[Hashable]]): node(s) to 
#                 add to the 'contents' attribute.

#         Raises:
#             ValueError: if 'item' already is in the stored tree or if 'parent'
#                 is not in the tree.
                            
#         """
#         if parent is None:
#             parent_node = self
#         else:
#             parent_node = self.get(item = parent)
#         if parent_node is None:
#             raise ValueError(
#                 f'Cannot add {item.name} because parent node {parent} is not '
#                 f'in the tree')
#         if isinstance(item, Sequence) and not isinstance(item, str):
#             for node in item:
#                 self.add(item = node)
#         elif item in self.nodes:
#             raise ValueError(
#                 f'Cannot add {item.name} because it is already in the tree')
#         else:
#             item.parent = parent_node
#             parent_node.contents.append(item)
#         return
    
#     def find(self, finder: Finder, **kwargs: Any) -> Optional[Hashable]:
#         """Finds first matching node in Tree using 'finder'.

#         Args:
#             finder (Callable[[Hashable], Optional[Hashable]]): 
#                 function or other callable that returns a node if it meets 
#                 certain criteria or otherwise returns None.
#             kwargs: keyword arguments to pass to 'finder' when examing each
#                 node.

#         Returns:
#             Optional[Hashable]: matching Node or None if no matching node 
#                 is found.
            
#         """                  
#         for node in self.nodes:
#             comparison = finder(self, **kwargs)
#             if comparison:
#                 return node
#         return None
            
#     def find_add(
#         self, 
#         finder: Finder, 
#         item: Hashable, 
#         **kwargs: Any) -> None:
#         """Finds first matching node in Tree using 'finder'.

#         Args:
#             finder (Callable[[Hashable], Optional[Hashable]]): 
#                 function or other callable that returns a node if it meets 
#                 certain criteria or otherwise returns None.
#             item (Hashable): node to add to the 'contents' attribute of 
#                 the first node that meets criteria in 'finder'.
#             kwargs: keyword arguments to pass to 'finder' when examing each
#                 node.

#         Raises:
#             ValueError: if no matching node is found by 'finder'.

#         Returns:
#             Optional[Hashable]: matching Node or None if no matching node 
#                 is found.
            
#         """  
#         node = self.find(finder = finder, **kwargs)
#         if node:
#             node.add(item = item)
#         else:
#             raise ValueError(
#                 'item could not be added because no matching node was found by '
#                 'finder')
#         return
    
#     def find_all(self, finder: Finder, **kwargs: Any) -> list[Hashable]:
#         """Finds all matching nodes in Tree using 'finder'.

#         Args:
#             finder (Callable[[Hashable], Optional[Hashable]]): 
#                 function or other callable that returns a node if it meets 
#                 certain criteria or otherwise returns None.
#             kwargs: keyword arguments to pass to 'finder' when examing each
#                 node.

#         Returns:
#             list[Hashable]: matching nodes or an empty list if no 
#                 matching node is found.
            
#         """              
#         found = []     
#         for node in self.nodes:
#             comparison = finder(self, **kwargs)
#             if comparison:
#                 found.append(node)
#         return found
            
#     def find_change(
#         self, 
#         finder: Finder, 
#         changer: Changer, 
#         **kwargs: Any) -> None:
#         """Finds matching nodes in Tree using 'finder' and applies 'changer'.

#         Args:
#             finder (Callable[[Hashable], Optional[Hashable]]): 
#                 function or other callable that returns a node if it meets 
#                 certain criteria or otherwise returns None.
#             changer (Callable[[Hashable], None]): function or other 
#                 callable that modifies the found node.
#             kwargs: keyword arguments to pass to 'finder' when examing each
#                 node.

#         Raises:
#             ValueError: if no matching node is found by 'finder'.
            
#         """  
#         nodes = self.find_all(finder = finder, **kwargs)
#         if nodes:
#             for node in nodes:
#                 changer(node)
#         else:
#             raise ValueError(
#                 'changer could not be applied because no matching node was '
#                 'found by finder')
#         return
    
#     def get(self, item: str) -> Optional[Hashable]:
#         """Finds first matching node in Tree match 'item'.

#         Args:
#             item (str): 

#         Returns:
#             Optional[Hashable]: matching Node or None if no matching node 
#                 is found.
            
#         """                  
#         for node in self.nodes:
#             if node.name == item:
#                 return node
#         return self.__missing__()
                                    
#     def walk(self, depth_first: bool = True) -> base.Path:
#         """Returns all paths in tree from 'start' to 'stop'.
        
#         Args:
#             depth_first (bool): whether to search through the stored tree depth-
#                 first (True) or breadth_first (False). Defaults to True.
                
#         """
#         if depth_first:
#             return depth_first_search(tree = self.contents)
#         else:
#             raise NotImplementedError(
#                 'breadth first search is not yet implemented')
#             # return breadth_first_search(tree = self.contents)

#     """ Dunder Methods """

#     def __add__(self, other: base.Graph) -> None:
#         """Adds 'other' to the stored tree using the 'append' method.

#         Args:
#             other (base.Graph): another Graph or supported
#                 raw data structure.
            
#         """
#         self.append(item = other)     
#         return 

#     def __radd__(self, other: base.Graph) -> None:
#         """Adds 'other' to the stored tree using the 'prepend' method.

#         Args:
#             other (base.Graph): another Graph or supported
#                 raw data structure.
            
#         """
#         self.prepend(item = other)     
#         return 

#     def __missing__(self) -> dict[str, Tree]:
#         """[summary]

#         Returns:
#             dict[str, Tree]: [description]
            
#         """
#         return {}
    
#     def __hash__(self) -> int:
#         """[summary]

#         Returns:
#             int: [description]
            
#         """
#         return hash(self.name)

#     def __eq__(self, other: Any) -> bool:
#         """[summary]

#         Args:
#             other (Any): [description]

#         Returns:
#             bool: [description]
            
#         """
#         if hasattr(other, 'name'):
#             return other.name == self.name
#         else:
#             return False
        
#     def __ne__(self, other: Any) -> bool:
#         """[summary]

#         Args:
#             other (Any): [description]

#         Returns:
#             bool: [description]
            
#         """
#         return not self.__eq__(other = other)


# def breadth_first_search(
#     tree: Tree, 
#     visited: Optional[list[Tree]] = None) -> base.Path:
#     """Returns a breadth first search path through 'tree'.

#     Args:
#         tree (Tree): tree to search.
#         visited (Optional[list[Tree]]): list of visited nodes. Defaults to None.

#     Returns:
#         base.Path: nodes in a path through 'tree'.
        
#     """         
#     visited = visited or []
#     if hasattr(tree, 'is_root') and tree.is_root:
#         visited.append(tree)
#     if hasattr(tree, 'children') and tree.children:
#         visited.extend(tree.children)
#         for child in tree.children:
#             visited.extend(breadth_first_search(tree = child, visited = visited))
#     return visited
                
                     
# def depth_first_search(
#     tree: Tree, 
#     visited: Optional[list[Tree]] = None) -> base.Path:
#     """Returns a depth first search path through 'tree'.

#     Args:
#         tree (Tree): tree to search.
#         visited (Optional[list[Tree]]): list of visited nodes. Defaults to None.

#     Returns:
#         base.Path: nodes in a path through 'tree'.
        
#     """  
#     visited = visited or []
#     visited.append(tree)
#     if hasattr(tree, 'children') and tree.children:
#         for child in tree.children:
#             visited.extend(depth_first_search(tree = child, visited = visited))
#     return visited

 
# @dataclasses.dataclass # type: ignore
# class Paths(sequence.Hybrid, base.Graph):
#     """Base class a collection of Path instances.
        
#     Args:
#         contents (MutableSequence[Hashable]): list of stored Path instances. 
#             Defaults to an empty list.

#     """
#     contents: MutableSequence[Path] = dataclasses.field(
#         default_factory = list)

#     """ Properties """

#     def endpoint(self) -> Path:
#         """Returns the endpoint of the stored composite object."""
#         return self.contents[list(self.contents.keys())[-1]]

#     def root(self) -> Path:
#         """Returns the root of the stored composite object."""
#         self.contents[list(self.contents.keys())[0]]
    
#     """ Public Methods """
  
#     def merge(item: base.Graph, *args: Any, **kwargs: Any) -> None:
#         """Combines 'item' with the stored composite object.

#         Args:
#             item (Graph): another Graph object to add to the stored 
#                 composite object.
                
#         """
#         pass

#     def walk(
#         self, 
#         start: Optional[Hashable] = None,
#         stop: Optional[Hashable] = None, 
#         path: Optional[Path] = None,
#         return_paths: bool = True, 
#         *args: Any, 
#         **kwargs: Any) -> Union[Path, Paths]:
#         """Returns path in the stored composite object from 'start' to 'stop'.
        
#         Args:
#             start (Optional[Hashable]): Hashable to start paths from. Defaults to None.
#                 If it is None, 'start' should be assigned to one of the roots
#                 of the base.
#             stop (Optional[Hashable]): Hashable to stop paths. Defaults to None. If it 
#                 is None, 'start' should be assigned to one of the roots of the 
#                 base.
#             path (Optional[Path]): a path from 'start' to 'stop'. 
#                 Defaults to None. This parameter is used by recursive methods 
#                 for determining a path.
#             return_paths (bool): whether to return a Paths instance 
#                 (True) or a Path instance (False). Defaults to True.

#         Returns:
#             Union[Path, Paths]: path(s) through the 
#                 Graph object. If multiple paths are possible and 
#                 'return_paths' is False, this method should return a 
#                 Path that includes all such paths appended to each other. If 
#                 multiple paths are possible and 'return_paths' is True, a 
#                 Paths instance with all of the paths should be returned. 
#                 Defaults to True.
                            
#         """
#         return self.items()
        
#     """ Dunder Methods """
        
#     @classmethod
#     def __instancecheck__(cls, instance: object) -> bool:
#         return check.is_paths(item = instance)
 


"""
base: base classes for composite data structures
Corey Rayburn Yung <coreyrayburnyung@gmail.com>
Copyright 2020-2022, Corey Rayburn Yung
License: Apache-2.0

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

Contents:
    Forms (object): stores all direct subclasses of Composite and Graph and 
        provides convenient classification and transformation methods.
    Composite (amos.Bunch, abc.ABC): base class for all composite data 
        structures.
    Graph (Composite, abc.ABC): base class for graphs.
    Edge (Sequence): base class for an edge in a graph. Many graphs will not
        require edge instances, but the class is made available for more complex 
        graphs and type checking.
    Node (Hashable): wrapper for items that can be stored in a composite data
        structure.
    classify: returns str name of the subtype a subtype of the passed item.
    transform: general subtype transformer that allows any base form of a 
        Composite or Graph to be changed to any other recognized form.
                 
To Do:
    Integrate Kinds system when it is finished.
    
"""
from __future__ import annotations
import abc
from collections.abc import Collection, Hashable, Sequence
import contextlib
import dataclasses
import inspect
from typing import Any, ClassVar, Optional, Type, Union

import amos 

from . import check
from . import workshop


""" Composite Form Registry """

@dataclasses.dataclass
class Forms(object):
    """Registry of composite data structures.
    
    Args:
        registry (amos.Dictionary): keys are names of composite data structure 
            forms and values are Composite and Graph subclasses.
                          
    """
    registry: ClassVar[amos.Dictionary] = amos.Dictionary()
    
    """ Public Methods """
    
    @classmethod
    def classify(cls, item: object) -> str:
        """Determines which form of composite that 'item' is.
        
        There is no difference between this classmethod and the 'classify'
        function.
        
        Args:
            item (object): object to classify.
            
        Returns:
            str: name of form that 'item' is.
            
        """
        return classify(item = item)
    
    @classmethod
    def register(
        cls, 
        item: Type[Composite], 
        name: Optional[str] = None) -> None:
        """Adds 'item' to 'registry'.
        
        The key assigned for storing 'item' is determined using the amos.namify
        method if 'name' is not passed.
        
        Args:
            item (Type[Composite]): class to register.
            name (Optional[str]): key to use for storing 'item'. Defaults to
                None.
            
        """
        name = name or amos.namify(item = item)
        cls.registry.add(item = {name: item})
        return
        
    @classmethod
    def transform(
        cls,
        item: Composite, 
        output: str, 
        raise_same_error: Optional[bool] = True) -> Composite:
        """General transform method that will call appropriate transformer.
        
        Unlike the 'transform' function, this method will return a Composite 
        wrapped in the form type stored in the Forms registry (as opposed to the 
        raw type). So, if 'output' is 'edges', this method will return an edge 
        list in the Edges class. In contrast, the 'transform' method will return 
        the structural type of an edge list without using the Edges class. The 
        rest of the logic is identical between the function and method.

        Args:
            item (Composite): Composite to transform.
            output (str): name of form to transform 'item' to.
            raise_same_error(Optional[bool]): whether to return an error if the 
                form of 'item' is the same as 'output'. If True, a ValueError 
                will be returned. If False, item will be return without any 
                change. Defaults to True.
        
        Raises:
            ValueError: if the form of 'item' is the same as 'output' and 
                'raise_same_error' is True.
                
        Returns:
            Composite: transformed composite data structure.
            
        """
        form = cls.classify(item = item)
        if form == output and raise_same_error:
            raise ValueError('The passed item and output are the same type')
        elif form == output:
            return item
        else:
            transformer = getattr(workshop, [f'{form}_to_{output}'])
            base = cls.registry[output]
            return base(contents = transformer(item = item))        

            
""" Base Classes for Composite Data Structures """
      
@dataclasses.dataclass
class Composite(abc.ABC):
    """Base class for composite data structures.
                                     
    Args:
         contents (Optional[Collection[Any]]): stored nodes or node labels. 
            Subclasses should narrow the type for contents based on the internal 
            storage format used.
                  
    """  
    contents: Optional[Collection[Any]] = None
    
    """ Initialization Methods """
    
    @classmethod
    def __init_subclass__(cls, *args: Any, **kwargs: Any):
        """Automatically registers subclass.."""
        # Because Composite will be used with mixins, it is important to call 
        # other '__init_subclass__' methods, if they exist.
        with contextlib.suppress(AttributeError):
            super().__init_subclass__(*args, **kwargs) # type: ignore
        # Adds a subclass to the Forms registry only if it is a direct subclass
        # of Composite.
        if Composite in cls.__bases__ and abc.ABC not in cls.__bases__:
            Forms.register(item = cls)
                          
    """ Public Methods """

    def add(self, item: Hashable, *args: Any, **kwargs: Any) -> None:
        """Adds node to the stored composite data structure.
                   
        Args:
            item (Hashable): node to add to the stored composite data structure.
            
        Raises:
            TypeError: if 'item' is not a node type.
            ValueError: if 'item' is already in the stored composite data 
                structure.
            
        """
        if not check.is_node(item = item):
            raise TypeError(f'{item} is not a node type')
        elif item in self.contents:
            raise ValueError(
                f'{item} is already in the composite data structure') 
        else:  
            self._add(item, *args, **kwargs)
        return
        
    def delete(self, item: Hashable, *args: Any, **kwargs: Any) -> None:
        """Deletes node from the stored composite data structure.

        Subclasses must provide their own specific methods for deleting a single
        node. The provided 'delete' method offers all of the error checking and
        the ability to delete multiple nodes at once. Subclasses just need to 
        provide the mechanism for deleting a single node without worrying about
        validation or error-checking.
                
        Args:
            item (Hashable): node to delete from 'contents'.
        
        Raises:
            KeyError: if 'item' is not in 'contents'.
            TypeError: if 'item' is not in 'contents'.
            
        """
        if not check.is_node(item = item):
            raise TypeError(f'{item} is not a node type')
        try:
            self._delete(item, *args **kwargs)
        except KeyError:
            raise KeyError(
                f'{item} does not exist in the composite data structure')
        return
    
    def merge(self, item: Composite, *args: Any, **kwargs: Any) -> None:
        """Adds 'item' to this Composite.

        This method is roughly equivalent to a dict.update, just adding 'item'
        to the existing stored composite data structure while maintaining its 
        structure.
        
        Args:
            item (base.Composite): another Composite to merge with
            
        Raises:
            TypeError: if 'item' is not compatible composite data structure 
                type.
            
        """
        if not check.is_composite(item = item):
            raise TypeError(f'{item} is not a compatible type') 
        else:
            self._merge(item, *args, **kwargs)
        return
        
    def subset(
        self, 
        include: Union[Hashable, Sequence[Hashable]] = None,
        exclude: Union[Hashable, Sequence[Hashable]] = None) -> Composite:
        """Returns a new Composite without a subset of 'contents'.
        
        All edges will be removed that include any nodes that are not part of
        the new composite data structure.
        
        Any extra attributes that are part of a Composite (or a subclass) should 
        be maintained in the returned composite data structure.

        Args:
            include (Union[Hashable, Sequence[Hashable]]): nodes or edges which 
                should be included in the new composite data structure.
            exclude (Union[Hashable, Sequence[Hashable]]): nodes or edges which 
                should not be included in the new composite data structure.

        Raises:
            ValueError: if include and exclude are none or if any item in
                include or exclude is not in the stored composite data 
                structure.

        Returns:
           Composite: with only selected nodes and edges.

        """
        if include is None and exclude is None:
            raise ValueError('Either include or exclude must not be None')
        if not all(i for i in include if i in self.contents):
            raise ValueError(
                'Some values in include are not in the composite data '
                'structure')
        if not all(i for i in exclude if i in self.contents):
            raise ValueError(
                'Some values in exclude are not in the composite data '
                'structure')
        return self._subset(include, exclude)
    
    """ Private Methods """
    
    def _add(self, item: Hashable, *args: Any, **kwargs: Any) -> None:
        """Adds node to the stored composite data structure.
 
        Subclasses must provide their own specific methods for adding a single
        node. The provided 'add' method offers all of the error checking and
        the ability to add multiple nodes at once. Subclasses just need to 
        provide the mechanism for adding a single node without worrying about
        validation or error-checking.
                   
        Args:
            item (Hashable): node to add to the stored composite data structure.
                
        """
        raise NotImplementedError
      
    def _delete(self, item: Hashable, *args: Any, **kwargs: Any) -> None:
        """Deletes node from the stored composite data structure.

        Subclasses must provide their own specific methods for deleting a single
        node. The provided 'delete' method offers all of the error checking and
        the ability to delete multiple nodes at once. Subclasses just need to 
        provide the mechanism for deleting a single node without worrying about
        validation or error-checking.
                
        Args:
            item (Hashable): node to delete from 'contents'.
            
        """
        raise NotImplementedError

    def _merge(self, item: Composite, *args: Any, **kwargs: Any) -> None:
        """Combines 'item' with the stored composite data structure.

        Subclasses must provide their own specific methods for merging with
        another composite data structure. The provided 'merge' method offers all 
        of the error checking. Subclasses just need to provide the mechanism for 
        merging ithout worrying about validation or error-checking.
        
        Args:
            item (Composite): another Composite object to add to the stored 
                composite data structure.
                
        """
        raise NotImplementedError
    
    def _subset(
        self, 
        include: Union[Hashable, Sequence[Hashable]] = None,
        exclude: Union[Hashable, Sequence[Hashable]] = None) -> Composite:
        """Returns a new Composite without a subset of 'contents'.

        Subclasses must provide their own specific methods for deleting a single
        edge. Subclasses just need to provide the mechanism for returning a
        subset without worrying about validation or error-checking.
        
        Args:
            include (Union[Hashable, Sequence[Hashable]]): nodes or edges which 
                should be included in the new composite data structure.
            exclude (Union[Hashable, Sequence[Hashable]]): nodes or edges which 
                should not be included in the new composite data structure.

        Returns:
           Graph: with only selected nodes and edges.
            
        """
        raise NotImplementedError  
                              
    """ Dunder Methods """
    
    @classmethod
    def __instancecheck__(cls, instance: object) -> bool:
        """Returns whether 'instance' meets criteria to be a subclass.

        Args:
            instance (object): item to test as an instance.

        Returns:
            bool: whether 'instance' meets criteria to be a subclass.
            
        """
        return check.is_composite(item = instance)
    
    
@dataclasses.dataclass
class Graph(Composite, abc.ABC):
    """Base class for holden graphs.
    
    Graph adds the requirements of '_connect' and '_disconnect' methods in 
    addition to the requirements of Composite.
    
    Args:
         contents (Optional[Collection[Any]]): stored nodes, node labels, edges,
            or edge labels. Subclasses should narrow the type for contents based
            on the internal storage format used.
                  
    """  
    contents: Optional[Collection[Any]] = None
    
    """ Initialization Methods """
    
    @classmethod
    def __init_subclass__(cls, *args: Any, **kwargs: Any):
        """Automatically registers subclass.."""
        # Because Graph will be used with mixins, it is important to call other 
        # '__init_subclass__' methods, if they exist.
        with contextlib.suppress(AttributeError):
            super().__init_subclass__(*args, **kwargs) # type: ignore
        # Adds a subclass to the Forms registry only if it is a direct subclass
        # of Graph.
        if Graph in cls.__bases__:
            Forms.register(item = cls)
                          
    """ Public Methods """

    def connect(self, item: Edge, *args: Any, **kwargs: Any) -> None:
        """Adds edge to the stored graph.
        
        Args:
            item (Edge): edge to add to the stored graph.
            
        Raises:
            ValueError: if the ends of the item are the same or if one of the
                edge ends does not currently exist in the stored graph.
            
        """
        if not check.is_edge(item = item):
            raise TypeError(f'{item} is not an edge type') 
        elif item[0] == item[1]:
            raise ValueError(
                'The starting point of an edge cannot be the same as the '
                'ending point')
        elif item[0] not in self:
            raise ValueError(f'{item[0]} is not in the graph')
        elif item[1] not in self:
            raise ValueError(f'{item[1]} is not in the graph')
        else:
            self._connect(item, *args, **kwargs)
        return                         

    def disconnect(self, item: Edge, *args: Any, **kwargs: Any) -> None:
        """Removes edge from the stored graph.
        
        Args:
            item (Edge): edge to delete from the stored graph.
            
        Raises:
            ValueError: if the edge does not exist in the stored graph.
            
        """
        if not check.is_edge(item = item):
            raise TypeError(f'{item} is not an edge type') 
        elif item[0] == item[1]:
            raise ValueError(
                'The starting point of an edge cannot be the same as the '
                'ending point')
        else:
            try:
                self._disconnect(item, *args, **kwargs)
            except (KeyError, ValueError):
                raise ValueError(
                    f'The edge ({item[0]}, {item[1]}) is not in the graph')
        return     

    """ Private Methods """

    def _connect(self, item: Edge, *args: Any, **kwargs: Any) -> None:
        """Adds edge to the stored graph.

        Subclasses must provide their own specific methods for adding a single
        edge. The provided 'connect' method offers all of the error checking and
        the ability to add multiple edges at once. Subclasses just need to 
        provide the mechanism for adding a single edge without worrying about
        validation or error-checking.
        
        Args:
            item (Edge): edge to add to the stored graph.
            
        """
        raise NotImplementedError

    def _disconnect(self, item: Edge, *args: Any, **kwargs: Any) -> None:
        """Removes edge from the stored graph.

        Subclasses must provide their own specific methods for deleting a single
        edge. The provided 'disconnect' method offers all of the error checking 
        and the ability to delete multiple edges at once. Subclasses just need 
        to provide the mechanism for deleting a single edge without worrying 
        about validation or error-checking.
        
        Args:
            item (Edge): edge to delete from the stored graph.
            
        """
        raise NotImplementedError  
                              
    """ Dunder Methods """
    
    @classmethod
    def __instancecheck__(cls, instance: object) -> bool:
        """Returns whether 'instance' meets criteria to be a subclass.

        Args:
            instance (object): item to test as an instance.

        Returns:
            bool: whether 'instance' meets criteria to be a subclass.
            
        """
        return check.is_graph(item = instance)

        
@dataclasses.dataclass(frozen = True, order = True)
class Edge(Sequence):
    """Base class for an edge in a graph structure.
        
    Edges are not required for most of the base graph classes in holden. But
    they can be used by subclasses of those base classes for more complex data
    structures.
    
    Args:
        start (Hashable): starting point for the edge.
        stop (Hashable): stopping point for the edge.
        
    """
    start: Hashable
    stop: Hashable

    """ Dunder Methods """
   
    def __getitem__(self, index: int) -> Hashable:
        """Allows Edge subclass to be accessed by index.
        
        Args:
            index (int): the number of the field in the dataclass based on 
                order.
        
        Raises:
            IndexError: if 'index' is greater than 1.
        
        Returns:
            Hashable: contents of field identified by 'index'.
                 
        """
        if index > 1:
            raise IndexError('Index out of bounds - edges are only two points')
        else:
            return getattr(self, dataclasses.fields(self)[index].name)
        
    @classmethod
    def __instancecheck__(cls, instance: object) -> bool:
        """Returns whether 'instance' meets criteria to be a subclass.

        Args:
            instance (object): item to test as an instance.

        Returns:
            bool: whether 'instance' meets criteria to be a subclass.
            
        """
        return check.is_edge(item = instance)
     
    def __len__(self) -> int:
        """Returns length of 2.
        
        Returns:
            int: 2
            
        """
        return 2


@dataclasses.dataclass
class Node(Hashable):
    """Vertex wrapper to provide hashability to any object.
    
    Node acts a basic wrapper for any item stored in a graph structure.
    
    Args:
        contents (Optional[Any]): any stored item(s). Defaults to None.
            
    """
    contents: Optional[Any] = None

    """ Initialization Methods """
    
    def __init_subclass__(cls, *args: Any, **kwargs: Any):
        """Forces subclasses to use the same hash methods as Node.
        
        This is necessary because dataclasses, by design, do not automatically 
        inherit the hash and equivalance dunder methods from their super 
        classes.
        
        """
        # Calls other '__init_subclass__' methods for parent and mixin classes.
        with contextlib.suppress(AttributeError):
            super().__init_subclass__(*args, **kwargs) # type: ignore
        # Copies hashing related methods to a subclass.
        cls.__hash__ = Node.__hash__ # type: ignore
        cls.__eq__ = Node.__eq__ # type: ignore
        cls.__ne__ = Node.__ne__ # type: ignore
                
    """ Dunder Methods """
    
    # @classmethod
    # def __subclasshook__(cls, subclass: Type[Any]) -> bool:
    #     """Returns whether 'subclass' is a virtual or real subclass.

    #     Args:
    #         subclass (Type[Any]): item to test as a subclass.

    #     Returns:
    #         bool: whether 'subclass' is a real or virtual subclass.
            
    #     """
    #     return check.is_node(item = subclass)
               
    # @classmethod
    # def __instancecheck__(cls, instance: object) -> bool:
    #     """Returns whether 'instance' meets criteria to be a subclass.

    #     Args:
    #         instance (object): item to test as an instance.

    #     Returns:
    #         bool: whether 'instance' meets criteria to be a subclass.
            
    #     """
    #     return check.is_node(item = instance)
    
    def __hash__(self) -> int:
        """Makes Node hashable so that it can be used as a key in a dict.

        Rather than using the object ID, this method prevents two Nodes with
        the same name from being used in a graph object that uses a dict as
        its base storage type.
        
        Returns:
            int: hashable of 'name'.
            
        """
        return hash(dataclasses.astuple(self))


""" Subtype Checker """

def classify(item: object) -> str:
    """Determines which form of graph that 'item' is.
    
    Args:
        item (object): object to classify.
        
    Returns:
        str: name of form that 'item' is.
        
    """
    # Chcecks for a matching parent clas in 'registry'.
    if inspect.isclass(item):
        subtype = item
    else:
        subtype = item.__class__       
    for name, form in Forms.registry.items():
        if issubclass(subtype, form):
            return name
    # Chaecks for matching raw form using functions in the 'check' module.
    for name in Forms.registry.keys():
        try:
            checker = getattr(check, f'is_{name}')
            if checker(item = item):
                return name
        except AttributeError:
            pass
    raise TypeError('The passed item is not a recognized Graph form')

""" Graph Transformer """

def transform(
    item: Graph, 
    output: str, 
    raise_same_error: Optional[bool] = True) -> Graph:
    """General transform function that will call appropriate transformer.

    Args:
        item Graph): Graph to transform.
        output (str): name of form to transform 'item' to.
        raise_same_error(Optional[bool]): whether to return an error if the form
            of 'item' is the same as 'output'. If True, a ValueError will be 
            returned. If False, item will be return without any change. Defaults
            to True.
    
    Raises:
        ValueError: if the form of 'item' is the same as 'output' and 
            'raise_same_error' is True.
            
    Returns:
        base.Graph: transformed graph.
        
    """
    form = Forms.classify(item = item)
    if form == output and raise_same_error:
        raise ValueError('The passed item and output are the same type')
    elif form == output:
        return item
    else:
        transformer = getattr(workshop, f'{form}_to_{output}')
        return transformer(item = item)



"""
traits: characteristics of graphs, edges, and nodes
Corey Rayburn Yung <coreyrayburnyung@gmail.com>
Copyright 2020-2022, Corey Rayburn Yung
License: Apache-2.0

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

Contents:
    Directed (abc.ABC): a directed graph with unweighted edges.
    Fungible (abc.ABC):
    Labeled (abc.ABC):    
    Storage (abc.ABC):
    Weighted (abc.ABC):
    
To Do:

    
"""
from __future__ import annotations
import abc
from collections.abc import Collection, Hashable
import contextlib
import dataclasses
from typing import Any, Optional, Type, TYPE_CHECKING, Union

import amos

from . import base

if TYPE_CHECKING:
    from . import composites
    from . import graphs
    
    
@dataclasses.dataclass
class Directed(abc.ABC):
    """Base class for directed graph data structures.
    
    Args:
        contents (Collection[Any]): stored collection of nodes and/or edges.
                                      
    """  
    contents: Collection[Any]
    
    """ Required Subclass Properties """
        
    @abc.abstractproperty
    def endpoint(self) -> Union[Hashable, Collection[Hashable]]:
        """Returns the endpoint(s) of the stored composite."""
        pass
 
    @abc.abstractproperty
    def root(self) -> Union[Hashable, Collection[Hashable]]:
        """Returns the root(s) of the stored composite."""
        pass
            
    """ Required Subclass Methods """
    
    @abc.abstractmethod
    def append(
        self, 
        item: Union[Hashable, base.Graph], 
        *args: Any, 
        **kwargs: Any) -> None:
        """Appends 'item' to the endpoint(s) of the stored composite.

        Args:
            item (Union[Hashable, base.Graph]): a Node or Graph to 
                add to the stored composite.
                
        """
        pass
    
    @abc.abstractmethod
    def prepend(
        self, 
        item: Union[Hashable, base.Graph], 
        *args: Any, 
        **kwargs: Any) -> None:
        """Prepends 'item' to the root(s) of the stored composite.

        Args:
            item (Union[Hashable, base.Graph]): a Node or Graph to 
                add to the stored composite.
                
        """
        pass
    
    @abc.abstractmethod
    def walk(
        self, 
        start: Optional[Hashable] = None,
        stop: Optional[Hashable] = None, 
        path: Optional[base.Path] = None,
        *args: Any, 
        **kwargs: Any) -> base.Path:
        """Returns path in the stored composite from 'start' to 'stop'.
        
        Args:
            start (Optional[Hashable]): Node to start paths from. 
                Defaults to None. If it is None, 'start' should be assigned to 
                'root'.
            stop (Optional[Hashable]): Node to stop paths at. 
                Defaults to None. If it is None, 'start' should be assigned to 
                'endpoint'.
            path (Optional[base.Path]): a path from 'start' to 'stop'. 
                Defaults to None. This parameter is used for recursively
                determining a path.

        Returns:
            base.Path: path(s) through the graph. 
                            
        """
        pass
    
    """ Dunder Methods """

    def __add__(self, other: base.Graph) -> None:
        """Adds 'other' to the stored composite using 'append'.

        Args:
            other (Union[base.Graph]): another graph to add to the current 
                one.
            
        """
        self.append(item = other)     
        return 

    def __radd__(self, other: base.Graph) -> None:
        """Adds 'other' to the stored composite using 'prepend'.

        Args:
            other (Union[base.Graph]): another graph to add to the current 
                one.
            
        """
        self.prepend(item = other)     
        return 
   
   
@dataclasses.dataclass # type: ignore
class Exportable(abc.ABC):
    """Mixin requirements for exporting graphs to other packages."""  
   
    """ Properties """

    @property
    def to_ascii(self) -> graphs.Adjacency:
        """Returns the stored composite as an Adjacency."""
        return base.transform(
            item = self, 
            output = 'adjacency', 
            raise_same_error = False)
        
    @property
    def to_dot(self) -> str:
        """Returns the stored composite in the graphviz dot format."""
        item = base.transform(
            item = self, 
            output = 'edges', 
            raise_same_error = False)
        
            

    
@dataclasses.dataclass # type: ignore
class Fungible(abc.ABC):
    """Mixin requirements for graphs that can be internally transformed."""  
   
    """ Properties """

    @property
    def adjacency(self) -> graphs.Adjacency:
        """Returns the stored composite as an Adjacency."""
        return base.transform(
            item = self, 
            output = 'adjacency', 
            raise_same_error = False)
        
    @property
    def edges(self) -> graphs.Edges:
        """Returns the stored composite as an Edges."""
        return base.transform(
            item = self, 
            output = 'edges', 
            raise_same_error = False)
           
    @property
    def matrix(self) -> graphs.Matrix:
        """Returns the stored composite as a Matrix."""
        return base.transform(
            item = self, 
            output = 'matrix', 
            raise_same_error = False)
     
    @property
    def parallel(self) -> composites.Parallel:
        """Returns the stored composite as a Parallel."""
        return base.transform(
            item = self, 
            output = 'parallel', 
            raise_same_error = False)
            
    @property
    def serial(self) -> composites.Serial:
        """Returns the stored composite as a Serial."""
        return base.transform(
            item = self, 
            output = 'serial', 
            raise_same_error = False)

    """ Class Methods """
    
    @classmethod
    def from_adjacency(cls, item: graphs.Adjacency) -> Fungible:
        """Creates a composite data structure from an Adjacency."""
        return cls(contents = base.transform(
            item = item, 
            output = base.classify(item = cls),
            raise_same_error = False))
    
    @classmethod
    def from_edges(cls, item: graphs.Edges) -> Fungible:
        """Creates a composite data structure from an Edges."""
        return cls(contents = base.transform(
            item = item, 
            output = base.classify(item = cls), 
            raise_same_error = False))
        
    @classmethod
    def from_matrix(cls, item: graphs.Matrix) -> Fungible:
        """Creates a composite data structure from a Matrix."""
        return cls(contents = base.transform(
            item = item, 
            output = base.classify(item = cls), 
            raise_same_error = False))
        
    @classmethod
    def from_parallel(cls, item: composites.Parallel) -> Fungible:
        """Creates a composite data structure from a Parallel."""
        return cls(contents = base.transform(
            item = item, 
            output = base.classify(item = cls), 
            raise_same_error = False)) 
        
    @classmethod
    def from_serial(cls, item: composites.Serial) -> Fungible:
        """Creates a composite data structure from a Serial."""
        return cls(contents = base.transform(
            item = item, 
            output = base.classify(item = cls), 
            raise_same_error = False))      
           
 
@dataclasses.dataclass
class Labeled(abc.ABC):
    """Mixin for labeling parts of a composite object. 

    Args:
        name (Optional[str]): designates the name of a class instance that is 
            used for internal and external referencing in a composite object.
            Defaults to None.
        contents (Optional[Any]): any stored item(s). Defaults to None.
            
    """
    name: Optional[str] = None
    contents: Optional[Any] = None
    
    """ Initialization Methods """

    def __post_init__(self) -> None:
        """Initializes instance."""
        # To support usage as a mixin, it is important to call other base class 
        # '__post_init__' methods, if they exist.
        with contextlib.suppress(AttributeError):
            super().__post_init__(*args, **kwargs) # type: ignore
        self.name = self.name or self._namify()

    """ Private Methods """
    
    def _namify(self) -> str:
        """Returns str name of an instance.
        
        By default, if 'contents' is None, 'none' will be returned. Otherwise, 
        amos.namify will be called based on the value of the 'contents'
        attribute and its return value will be returned. 
        
        For different naming rules, subclasses should override this method, 
        which is automatically called when an instance is initialized.
        
        Returns:
            str: str label for part of a composite data structute.
            
        """
        if self.contents is None:
            return 'none'
        else:
            return amos.namify(item = self.contents)
                               
    """ Dunder Methods """
    
    def __hash__(self) -> int:
        """Makes Node hashable based on 'name.'
        
        Returns:
            int: hashable of 'name'.
            
        """
        return hash(self.name)

    def __eq__(self, other: object) -> bool:
        """Determines equality based on 'name' attribute.

        Args:
            other (object): other object to test for equivalance.
            
        Returns:
            bool: whether 'name' is the same as 'other.name'.
            
        """
        try:
            return str(self.name) == str(other.name) # type: ignore
        except AttributeError:
            return str(self.name) == other

    # def __ne__(self, other: object) -> bool:
    #     """Determines inequality based on 'name' attribute.

    #     Args:
    #         other (object): other object to test for equivalance.
           
    #     Returns:
    #         bool: whether 'name' is not the same as 'other.name'.
            
    #     """
    #     return not(self == other)
 

@dataclasses.dataclass
class Storage(abc.ABC):
    """Mixin for storage of nodes in a Library with the composite object.
    
    Args:
        contents (Collection[Any]): stored collection of nodes and/or edges.
                                      
    """  
    contents: Collection[Any]
    nodes: amos.Library = dataclasses.field(default_factory = amos.Library)
 

@dataclasses.dataclass
class Weighted(abc.ABC):
    """Mixin for weighted nodes.
    
    Args:
        weight (Optional[float]): the weight of the object. Defaults to 1.0.
                  
    """  
    weight: Optional[float] = 1.0   

    """ Dunder Methods """  
        
    def __len__(self) -> float:
        """Returns 'weight'.
        
        Returns:
            float: weight of the edge.
            
        """
        return self.weight


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['holden']

package_data = \
{'': ['*']}

install_requires = \
['amos>=0.1.10,<0.2.0', 'miller>=0.1.2,<0.2.0']

setup_kwargs = {
    'name': 'holden',
    'version': '0.1.5',
    'description': 'lightweight, accessible, flexible composite data structures',
    'long_description': '[![made-with-python](https://img.shields.io/badge/Made%20with-Python-1f425f.svg)](https://www.python.org/) [![PyPI Latest Release](https://img.shields.io/pypi/v/holden.svg)](https://pypi.org/project/holden/) [![License](https://img.shields.io/badge/License-Apache_2.0-blue.svg)](https://opensource.org/licenses/Apache-2.0) [![Documentation Status](https://readthedocs.org/projects/holden/badge/?version=latest)](http://holden.readthedocs.io/?badge=latest)\n\n<p align="center">\n<img src="https://media.giphy.com/media/3ornjRyce6SukW8INi/giphy.gif" />\n</p>\n\nThis package is named after the Roci\'s captain in *The Expanse*, James Holden, who was adept at furling his brow and recognizing connections. In a similar vein, **holden** offers users easy-to-use composite data structures without the overhead or complexity of larger graph packages. The included graphs are built for basic workflow design or analysis of conditional relationships. They are not designed for big data network analysis or similar large-scale projects (although nothing prevents you from using them in that manner). Rather, the goal of **holden** is to provide lightweight, turnkey, extensible composite data structures without all of the stuff you don\'t need in packages like [networkx](https://github.com/networkx/networkx). **holden** serves as the base for my [chrisjen](https://github.com/WithPrecedent/chrisjen) workflow package (similarly named for a character from The Expanse), but I have made **holden** available separately for easier integration into other uses.\n\n## Simple\n\nThe basic building blocks provided are:\n* `Composite`: the abstract base class for all types of a composite data structures\n* `Graph`: subclass of Composite and the base class for all graph data structures\n* `Edge`: an optional edge class which can be treated as a drop-in tuple replacement or extended for greater functionality\n* `Node`: an optional vertex class which provides universal hashability and some other convenient functions\n* `Forms`: a dictionary that automatically stores all direct Compisite subclasses to allow flexible subtype checking of and transformation between composite subtypes using its `classify` and `transform` methods\n\nOut of the box, Graph has several subtypes with varying internal storage formats:\n* `Adjacency`: an adjacency list using a `dict(Node, set(Node))` structure\n* `Matrix`: an adjacency matrix that uses a `list[list[float | int]]` for mapping edges and a separate `list[str]` attribute that corresponds to the list of lists matrix\n* `Edges`: an edge list structure that uses a `list[tuple[Node, Node]]` format\n  \nYou can use **holden** without any regard to what is going on inside the graph. The methods and properties are the same regardless of which internal format is used. But the different forms are provided in case you want to utilize the advantages or avoid certain drawbacks of a particular form. Unless you want to design a different graph form, you should design subclasses to inherit from one of the\nincluded forms and add mixins to expand functionality.\n\n## Flexible\n\n Various traits can be added to graphs, nodes, and edges as mixins including:\n* Weighted edges (`Weighted`)\n* Abilty to create a graph from or convert any graph to any recognized form using properties with consistent syntax (`Fungible`)\n* Directed graphs (`Directed`)\n* Automatically names objects if a name is not passed (`Labeled`)\n* Ability to store node data internally for easy reuse separate from the graph structure (`Storage`)\n\n**holden** provides transformation methods between all of the internal storage forms as well as functions to convert graphs into a set of paths (`Parallel`) or a single path (`Serial`). The transformation methods can be used as class properties or with functions using an easy-to-understand naming convention (e.g., adjacency_to_edges or edges_to_parallel).\n\n**holden**\'s framework supports a wide range of coding styles. You can create complex multiple inheritance structures with mixins galore or simpler, compositional objects. Even though the data structures are necessarily object-oriented, all of the tools to modify them are also available as functions, for those who prefer a more functional approaching to programming.\n\nThe package also uses structural subtyping that allows raw forms of the supported composite subtypes to be used and recognized as the same forms for which **holden** includes classes. So, for example, the is_adjacency function will recognize any object with a dict(Node, set(Node)) structure and isinstance(item, **holden**.Adjacency) will similarly return True for a raw adjacency list.\n\n## Contributing\n\nThe project is highly internally documented so that users and developers can make **holden** work with their projects. It is designed for Python coders at all levels. Beginners should be able to follow the readable code and internal documentation to understand how it works. More advanced users should find complex and tricky problems addressed through efficient code.\n\nI hope you find **holden** useful and feel free to contribute, leave suggestions, or report bugs.\n\n<p align="center">\n<img src="https://media.giphy.com/media/3oKIPwyf0EBAGnAkWk/giphy.gif" />\n</p>\n\n## Similar Projects\n* [networkx](https://github.com/networkx/networkx): the market leader for python graphs. Offers greater flexibility and extensibility at the cost of substantial overhead.',
    'author': 'Corey Rayburn Yung',
    'author_email': 'coreyrayburnyung@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/WithPrecedent/holden',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
